/*
 * Copyright (C) 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "include.h"
#include "router.h"
#include "bgp.h"
#include "task.h"
#include "aspath.h"
#include "bgp_var.h"

static int bgp_read __P((struct rpcb *, int, int));

/*
 *    bgp_input()
 *             called by  main_listen_accept()  only.
 */
void
bgp_input(struct rpcb *bnp) {

  int            len;        /* by read() */
  u_int16_t      length;     /* by msg    */
  struct bgphdr *bh;


  char        in6txt[INET6_ADDRSTRLEN];
#ifdef DEBUG
  char      myin6txt[INET6_ADDRSTRLEN];
#endif

  extern byte   inpkt[];
  extern fd_set currentmask;

#ifdef DEBUG
  extern char *bgp_msgstr[], *bgp_statestr[];
#endif

  memset(inpkt, 0, BGPMAXPACKETSIZE);
  FD_CLR(bnp->rp_socket, &currentmask); /* currentmask is most meaningful */


  if (bnp->rp_socket == -1)             /* in main_listen_accept() loop. */
    fatal("<bgp_input>: invalid socket");

/*
   One method that can be used in this situation is to first try to read
   just the message header. For the KEEPALIVE message type, this is a
   complete message; for other message types, the header should first be
   verified, in particular the total length. If all checks are
   successful, the specified length, minus the size of the message
   header is the amount of data left to read.  [draft-ietf-idr-bgp4-06.txt]
*/
#if 0
  if ((len = read(bnp->rp_socket, inpkt, BGP_HEADER_LEN)) < 0) {
    syslog(LOG_ERR, "<bgp_input>: read from peer %s (%s AS %d) failed",
	   inet_ntop(AF_INET6, &bnp->rp_addr.sin6_addr,
		     in6txt, INET6_ADDRSTRLEN),
	   ((bnp->rp_mode & BGPO_IGP) ? "Internal" : "External"),
	   (int)bnp->rp_as);
    switch (errno) {
    case EINTR: case ECONNRESET: case ENOTCONN:
      bgp_cease(bnp);
      return;
      break;
    default:
      fatal("<bgp_input>: read() failed");
      break;
    }
  }
#else
  if ((len = bgp_read(bnp, 0, BGP_HEADER_LEN)) < 0)
	  return;
#endif

  if (len < BGP_HEADER_LEN) {
      bgp_notify(bnp, BGP_ERR_HEADER, BGP_ERRHDR_LENGTH, len, inpkt);
      return;
  }

  bh = (struct bgphdr *)inpkt;

  /* Length (2-octet) */
  length = ntohs(bh->bh_length);

  if (length < BGP_HEADER_LEN || length > BGPMAXPACKETSIZE) {
      bgp_notify(bnp, BGP_ERR_HEADER, BGP_ERRHDR_LENGTH,
		 2, (byte *)&bh->bh_length);
      return;
  }


  /* read remaining data */
#if 0
  if ((len = read(bnp->rp_socket, inpkt + BGP_HEADER_LEN,
		  length - BGP_HEADER_LEN)) < 0) {
    syslog(LOG_ERR, "<bgp_input>: read from peer %s (%s AS %d) failed",
	   inet_ntop(AF_INET6, &bnp->rp_addr.sin6_addr,
		     in6txt, INET6_ADDRSTRLEN),
	   ((bnp->rp_mode & BGPO_IGP) ? "Internal" : "External"),
	   (int)bnp->rp_as);
    switch (errno) {
    case EINTR: case ECONNRESET: case ENOTCONN:
      bgp_cease(bnp);
      return;
      break;
    default:
      fatal("<bgp_input>: read() failed");
      break;
    }
  }
#else
  if ((len = bgp_read(bnp, BGP_HEADER_LEN, length - BGP_HEADER_LEN)) < 0)
	  return;
#endif
  len += BGP_HEADER_LEN;



#ifdef DEBUG
  syslog(LOG_DEBUG,
	 "BGP+ RECV %s+%d -> %s+%d",
	 inet_ntop(AF_INET6, &bnp->rp_addr.sin6_addr,
		   in6txt, INET6_ADDRSTRLEN),
	 ntohs(bnp->rp_addr.sin6_port),
	 inet_ntop(AF_INET6, &bnp->rp_myaddr.sin6_addr,
		   myin6txt, INET6_ADDRSTRLEN),
	 ntohs(bnp->rp_myaddr.sin6_port));

  if (BGP_TYPE_VALID(bh->bh_type))
    syslog(LOG_DEBUG,
	   "BGP+ RECV message type %d (%s) length %d, state=%s",
	   bh->bh_type,
	   bgp_msgstr[bh->bh_type],
	   len,
	   bgp_statestr[bnp->rp_state]);
#endif


  switch (bh->bh_type) {

  case BGP_OPEN:
    if ((length != len) || (len < BGP_HEADER_LEN + 10)) {
      /* Bad Message Length */
      bgp_notify(bnp, BGP_ERR_HEADER, BGP_ERRHDR_LENGTH,
		 2, (byte *)&bh->bh_length);
      return;
    }
    bgp_process_open(bnp);
    break;


  case BGP_UPDATE:
    if ((length != len) || (len < BGP_HEADER_LEN + 4 )) {
      /* Bad Message Length */
      bgp_notify(bnp, BGP_ERR_HEADER, BGP_ERRHDR_LENGTH,
		 2, (byte *)&bh->bh_length);
      return;
    }
    bgp_process_update(bnp);
    break;


  case BGP_NOTIFY:
    bgp_process_notification(bnp);
    break;


  case BGP_KEEPALIVE:
    if ((length != len) || (len != BGP_HEADER_LEN)) {
      /* Bad Message Length */
      bgp_notify(bnp, BGP_ERR_HEADER, BGP_ERRHDR_LENGTH,
		 2, (byte *)&bh->bh_length);
      return;
    }
    bgp_process_keepalive(bnp);
    break;


  default:
    /* If the Type field of the message header is not recognized, then the
       Error Subcode is set to Bad Message Type.  The Data field contains
       the erroneous Type field. */
    bgp_notify(bnp, BGP_ERR_HEADER, BGP_ERRHDR_TYPE, 1, &bh->bh_type);
    return;
    break;
  }
} /* End of bgp_input() */

/*
 * Read specified length data from a bgp socket. If the whole data aren't
 * read by the first call of read(), the function calls the system call
 * until it gets enough data.
 * XXX: This function may cause blocking for read() call.
 */
static int
bgp_read(struct rpcb *bnp, int off, int len)
{
	int cc, resid = len;
	extern byte inpkt[];
	char in6txt[INET6_ADDRSTRLEN];
	byte *buf = inpkt + off;

	while(resid) {
		cc = read(bnp->rp_socket, buf, resid);
		if (cc == 0) {
			/* This would occur when the peer close the connection */
			bgp_cease(bnp);
			return(-1);
		}
		if (cc < 0) {
			syslog(LOG_ERR,
			       "%s: read from peer %s (%s AS %d) failed",
			       __FUNCTION__, 
			       inet_ntop(AF_INET6, &bnp->rp_addr.sin6_addr,
					 in6txt, INET6_ADDRSTRLEN),
			       ((bnp->rp_mode & BGPO_IGP) ?
				"Internal" : "External"),
			       (int)bnp->rp_as);
			switch (errno) {
			 case EINTR: case ECONNRESET: case ENOTCONN:
				 bgp_cease(bnp);
				 return(-1);
				 break;
			 default:
				 fatal("<bgp_read>: read() failed");
				 break;
			}
		}
		resid -= cc;
		buf += cc;
	}

	return(len);
}
