/*
 * Copyright (C) 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "include.h"
#include "router.h"
#include "bgp.h"
#include "task.h"
#include "rt_table.h"
#include "aspath.h"
#include "bgp_var.h"
#include "in6.h"
#include "ripng.h"
#include "ripng_var.h"

/*
 *    rpcblookup()
 *      RETURN VALUES:   a Pointer:  found
 *                       NULL     : not found
 *
 *      DESCRIPTION:     find a router by ID.
 */
struct rpcb *
rpcblookup(head, id)
     struct rpcb *head;  /* list */
     u_int32_t    id;
{
  struct rpcb *rp;
  
  if (id == 0)
    return NULL;

  rp = head;
  while(rp) {
    if (rp->rp_id == id)
      return rp;

    if ((rp = rp->rp_next) == head)
      break;
  }
  return NULL;
}


/*
 *    find_rtp()
 */
struct rtproto *
find_rtp(key, base)
     struct rtproto *key;
     struct rtproto *base;
{ 
  struct rtproto *rtp;

  if ((key == NULL) ||
      ((rtp = base) == NULL))
    return NULL;

  while(rtp) {
    if (rtp->rtp_type == key->rtp_type)
      switch (key->rtp_type) {
      case RTPROTO_IF:
	if (key->rtp_if == rtp->rtp_if)
	  return rtp;	
	break;
      case RTPROTO_BGP:
	if (key->rtp_bgp->rp_mode & BGPO_IGP) { /* IBGP, compare BGP-ID */
	  if (key->rtp_bgp->rp_id &&
	      key->rtp_bgp->rp_id == rtp->rtp_bgp->rp_id)
	    return rtp;
	} else {                                  /* EBGP, compare AS num */
	  if (key->rtp_bgp->rp_as == rtp->rtp_bgp->rp_as)
	    return rtp;
	}
	break;
      case RTPROTO_RIP:
	if (rtp->rtp_rip == key->rtp_rip)
	  return rtp;
	break;
      default:
	fatal("<find_rtp>: BUG !");
	break;
      }

    if ((rtp = rtp->rtp_next) == base)
      return NULL;
  }

  return NULL; /* NOT REACHED */ 
}



/*
 *   propagate()
 */
void
propagate(rte)
     struct rt_entry *rte;       /* list               */
{
  struct rtproto  *srcrtp;    /* rtproto (struct rpcb), which RTEs got  */
  struct rt_entry *last;      /* (1998/05/29) */
  extern byte      bgpyes;
  extern byte      ripyes;

#ifdef DEBUG
  syslog(LOG_DEBUG, "<%s>: Invoked.", __FUNCTION__);
#endif

  if (rte == NULL) {
#ifdef DEBUG
    syslog(LOG_DEBUG, "<%s>: Nothing to do.", __FUNCTION__);
#endif
    return;
  }

  srcrtp = &rte->rt_proto;

  if (bgpyes) {
    struct rpcb        *bnp;              /* Propagated to... */
    extern struct rpcb *bgb;

    bnp = bgb;
    while(bnp) {

      if (bnp->rp_state == BGPSTATE_ESTABLISHED &&
	  find_rtp(srcrtp , bnp->rp_adj_ribs_out)) {
	struct rt_entry *irte;

	irte = rte;  /* iw97 */
	while (irte) { /* pointer maybe shifted */      /* (1998/05/29) */

#define BREAK_NEXT_BNP { bnp = bnp->rp_next ; break; }
	  if ((last = bgp_send_withdrawn(bnp, irte)) == NULL) {
	    bnp = bnp->rp_next;
	    bgp_cease(bnp->rp_prev);
	    break;
	  }

	  if ((irte != rte) && (last == rte)) {
	    BREAK_NEXT_BNP;
	  }
	  if ((irte = last->rt_next) == rte) {
	    BREAK_NEXT_BNP;
	  }
	} /* while(irte) */

      } else
	bnp = bnp->rp_next;

      if (bnp == bgb)
	break;
    } /* while(bnp) */
#undef BREAK_NEXT_BNP
  }

  if (ripyes) {
    struct ripif        *ripif;        /* Propagated to... */
    struct in6_pktinfo   spktinfo;
    struct riphdr       *rp;  
    int                  nn;

    extern struct sockaddr_in6  ripsin; /* ff02::9  */
    extern struct ripif        *ripifs;
    extern byte                 rippkt[], ripbuf[];

    rp = (struct riphdr *)rippkt;    /* outgoing RIPng header   */
    rp->riph_cmd   = RIPNGCMD_RESPONSE;
    rp->riph_vers  = RIPNG_VERSION;
    rp->riph_zero2 = 0;


    ripif = ripifs;
    while(ripif) {
      if (!(ripif->rip_mode & IFS_NORIPOUT) &&
	  find_rtp(srcrtp, ripif->rip_adj_ribs_out) &&
	  ripif != srcrtp->rtp_rip) {  /* split holizon */
	int mm;  /* current */
	int done = 0;


	nn = rip_make_data(rte, NULL);  /* withdrawning: (1998/06/11) */


	spktinfo.ipi6_addr    = ripif->rip_ife->ifi_laddr;  /* copy */
	spktinfo.ipi6_ifindex = ripif->rip_ife->ifi_ifn->if_index;


	while(1) {
	  mm = MIN(nn - done, RIPNG_MAXRTES);
	  memcpy(&rippkt[sizeof(struct riphdr)],
		 &ripbuf[sizeof(struct riphdr) + done*sizeof(struct ripinfo6)],
		 mm * sizeof(struct ripinfo6));

	  rip_sendmsg(&ripsin,
		      &spktinfo,
		      sizeof(struct riphdr) + mm * sizeof(struct ripinfo6));

	  done += mm;
	  if (done == nn) break;
	}
      }
      if ((ripif = ripif->rip_next) == ripifs)
	break;
    }
  }

#ifdef DEBUG
  syslog(LOG_DEBUG, "<%s>: executed.", __FUNCTION__);
#endif
  
}


/*
 *   redistribute()
 *     DESCRIPTION: Redistribute to IGP first,
 *                    then, Redistribute to iBGP. see [rfc1772]
 */
void
redistribute(rte)
     struct rt_entry *rte;   /* Redistribute RTEs                         */
{
  struct rtproto *srcrtp;    /* rtproto (struct rpcb), which new RTEs got */

  extern byte     bgpyes;
  extern byte     ripyes;

#ifdef DEBUG
  syslog(LOG_DEBUG, "<%s>: invoked.", __FUNCTION__);
#endif
 
  if (rte)
    srcrtp = &rte->rt_proto;
  else {
#ifdef DEBUG
    syslog(LOG_DEBUG, "<%s>: Nothing to do.", __FUNCTION__);
#endif
    return;
  }

  if (ripyes) {
    struct ripif        *ripif;   /* Redistributed to...  */
    struct in6_pktinfo   spktinfo;
    struct riphdr       *rp;  
    int                  nn;

    extern struct sockaddr_in6  ripsin; /* ff02::9  */
    extern struct ripif        *ripifs;
    extern byte                 rippkt[], ripbuf[];

    rp = (struct riphdr *)rippkt;    /* outgoing RIPng header   */
    rp->riph_cmd   = RIPNGCMD_RESPONSE;
    rp->riph_vers  = RIPNG_VERSION;
    rp->riph_zero2 = 0;


    ripif = ripifs;
    while(ripif) {

      if (!(ripif->rip_mode & IFS_NORIPOUT) &&
	  find_rtp(srcrtp , ripif->rip_adj_ribs_out) &&
	  ripif != srcrtp->rtp_rip) {  /* split holizon */
	int mm;  /* current */
	int done = 0;

	aggr_flush();                     /* (1998/06/11) */
	nn = rip_make_data(rte, ripif);


	spktinfo.ipi6_addr    = ripif->rip_ife->ifi_laddr;  /* copy */
	spktinfo.ipi6_ifindex = ripif->rip_ife->ifi_ifn->if_index;

	while(1) {
	  mm = MIN(nn - done, RIPNG_MAXRTES);
	  memcpy(&rippkt[sizeof(struct riphdr)],
		 &ripbuf[sizeof(struct riphdr) + done*sizeof(struct ripinfo6)],
		 mm * sizeof(struct ripinfo6));

	  rip_sendmsg(&ripsin,
		      &spktinfo,
		      sizeof(struct riphdr) + mm * sizeof(struct ripinfo6));

	  done += mm;
	  if (done == nn) break;
	}
      }
      if ((ripif = ripif->rip_next) == ripifs)
	break;
    }
  }

  if (bgpyes) {
    struct rpcb        *bnp;         /* Redistributed to...  */
    extern struct rpcb *bgb;
    bnp    = bgb;
    while(bnp) {
      if (bnp->rp_state == BGPSTATE_ESTABLISHED &&
	  find_rtp(srcrtp, bnp->rp_adj_ribs_out)) {
	aggr_flush();
	if (bgp_send_update(bnp, rte) == NULL) { /* msg length short enough */
	  bnp = bnp->rp_next;
	  bgp_cease(bnp->rp_prev);
	  bnp = bnp->rp_prev;
	}
      }
      if ((bnp = bnp->rp_next) == bgb)
	break;
    }
  }

#ifdef DEBUG
  syslog(LOG_DEBUG, "<%s>: executed.", __FUNCTION__);
#endif

}

