/*
 * Copyright (C) 1997 and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * Control command for kernel function of User level translator
 *  from IPv6 to IPv4.
 * Usage: faith [-de] [-p prefix]
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <err.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/sysctl.h>
#include <sys/resource.h>

#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

int pflag;

static void
usage(void)
{
	(void)fprintf(stderr, "usage: faith [-de] [-p prefix]\n");
	exit(1);
}

void
faith_disp(void)
{
	char prefix_str[INET6_ADDRSTRLEN];
	struct in6_addr prefix = IN6ADDR_ANY_INIT;
	int keepfaith = 0, mib[4], size;

	mib[0] = CTL_NET;
	mib[1] = PF_INET6;
	mib[2] = IPPROTO_IPV6;
	mib[3] = IPV6CTL_FAITH_PREFIX;

	size = sizeof(struct in6_addr);
	if (sysctl(mib, 4, &prefix, &size, NULL, 0) < 0) {
		err(1, "sysctl(net.inet6.ip6.faith_prefix)");
		/*NOTREACHED*/
	}
	inet_ntop(AF_INET6, &prefix, prefix_str, sizeof(prefix_str) - 1);
	printf("faith_prefix: %s\n", prefix_str);

	mib[3] = IPV6CTL_KEEPFAITH;
	size = sizeof(int);
	if (sysctl(mib, 4, &keepfaith, &size, NULL, 0) < 0) {
		err(1, "sysctl(net.inet6.ip6.keepfaith)");
		/*NOTREACHED*/
	}
	printf("faith: %s\n", keepfaith ? "on" : "off");

	return;
}

int
main(int argc, char **argv)
{
	int ch;
	int mib[4], size;
	struct in6_addr ip6_faith_prefix = IN6ADDR_ANY_INIT;
	int faithmode = -1;
	char *prefix = NULL;

	while ((ch = getopt(argc, argv, "dep:")) !=  EOF) {
		switch (ch) {
		case 'd':	/*disable*/
			faithmode = 0;
			break;
		case 'e':	/*enable*/
			faithmode = 1;
			break;
		case 'p':	/*prefix*/
			prefix = optarg;
			break;
		default:
			usage();
			/*NOTREACHED*/
		}
	}
	argc -= optind;
	argv += optind;

#if 0
	/* backward compatibility mode*/
	if (faithmode < 0 && !prefix) {
		if (2 < argc) {
			usage();
			/*NOTREACHED*/
		}
		while (argc) {
			if (strcmp(*argv, "on") == 0)
				faithmode = 1;
			else if (strcmp(*argv, "off") == 0)
				faithmode = 0;
			else {
				prefix = *argv;
				faithmode = 1;
			}
			argc--;
			argv++;
		}
	}
#endif

	if (argc) {
		usage();
		/*NOTREACHED*/
	}

	if (prefix) {
		if (inet_pton(AF_INET6, prefix, &ip6_faith_prefix) != 1) {
			errx(1, "invalid prefix \"%s\"", prefix);
			/*NOTREACHED*/
		}

#if 1
		/* sanity check */
		if (ip6_faith_prefix.s6_addr32[3]) {
			errx(1, "invalid prefix \"%s\" (not a prefix)", prefix);
			/*NOTREACHED*/
		}
#endif

		mib[0] = CTL_NET;
		mib[1] = PF_INET6;
		mib[2] = IPPROTO_IPV6;
		mib[3] = IPV6CTL_FAITH_PREFIX;
		size = sizeof(struct in6_addr);
		if (sysctl(mib, 4, 0, 0, &ip6_faith_prefix, size) < 0) {
			err(1, "sysctl(net.inet6.ip6.faith_prefix)");
			/*NOTREACHED*/
		}
	}
	if (faithmode == 0 || faithmode == 1) {
		mib[0] = CTL_NET;
		mib[1] = PF_INET6;
		mib[2] = IPPROTO_IPV6;
		mib[3] = IPV6CTL_KEEPFAITH;
		size = sizeof(int);
		if (sysctl(mib, 4, 0, 0, &faithmode, size) < 0) {
			err(1, "sysctl(net.inet6.ip6.keepfaith)");
			/*NOTREACHED*/
		}
	}
	faith_disp();
	exit(0);
}
