/*
 * Copyright (C) 1997 and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * User level translator from IPv6 to IPv4.
 *
 * Usage: faithd [<port> <progpath> <arg1(progname)> <arg2> ...]
 *   e.g. faithd telnet /usr/local/v6/sbin/telnetd telnetd
 */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/sysctl.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/ioctl.h>

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <fcntl.h>
#include <termios.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#include "faithd.h"

char *serverpath = NULL;
char *serverarg[MAXARGV + 1];

void
main(int argc, char *argv[])
{
	struct addrinfo hints, *res;
	struct sockaddr_in6 srcaddr6;
	char logname[BUFSIZ], *faithdname, *service;
	int s_wld, s_src, err, i, serverargc, on = 1, child_pid;

	/*
	 * Initializing stuff
	 */

	faithdname = strrchr(argv[0], '/');
	if (faithdname)
		faithdname++;
	else
		faithdname = argv[0];
	    
	switch (argc) {
	case 1:
		serverpath = DEFAULT_PATH;
		serverarg[0] = DEFAULT_NAME;
		serverarg[1] = NULL;
		service = DEFAULT_PORT_NAME;
		break;
	default:
		serverargc = argc - NUMARG;
		if (serverargc > MAXARGV)
			exit_error("too many augments");

		serverpath = malloc(strlen(argv[NUMPRG]));
		strcpy(serverpath, argv[NUMPRG]);
		for (i = 0; i < serverargc; i++) {
			serverarg[i] = malloc(strlen(argv[i + NUMARG]));
			strcpy(serverarg[i], argv[i + NUMARG]);
		}
		serverarg[i] = NULL;
		/* fall throuth */
	case 2:	/* no local service */
		service = argv[NUMPRT];
		break;
	}

	/*
	 * Opening wild card socket for this service.
	 */

	memset(&hints, 0, sizeof(hints));
	hints.ai_flags = AI_PASSIVE;
	hints.ai_family = PF_INET6;
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = 0;
	err = getaddrinfo(NULL, service, &hints, &res);
	if (err) 
		exit_error("gaddrinfo: %s", gai_strerror(err));

	s_wld = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	if (s_wld == -1)
		exit_error("socket: %s", ERRSTR);

	err = setsockopt(s_wld, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
	if (err == -1)
		exit_error("setsockopt: %s", ERRSTR);
	
	err = setsockopt(s_wld, SOL_SOCKET, SO_OOBINLINE, &on, sizeof(on));
	if (err == -1)
		exit_error("setsockopt: %s", ERRSTR);

	err = bind(s_wld, (struct sockaddr *)res->ai_addr, res->ai_addrlen);
	if (err == -1)
		exit_error("bind: %s", ERRSTR);

	err = listen(s_wld, 5);
	if (err == -1)
		exit_error("listen: %s", ERRSTR);

	/*
	 * Everything is OK.
	 */

	start_daemon();

	sprintf(logname, "%s(%s)", faithdname, service);
	openlog(logname, LOG_PID | LOG_NOWAIT, LOG_DAEMON);
	syslog(LOG_INFO, "Staring faith daemon for %s port", service);

	/*
	 * Wait, accept, fork, faith....
	 */

	for (;;) {
		int len = sizeof(srcaddr6);

		s_src = accept(s_wld, (struct sockaddr *)&srcaddr6, &len);
		if (s_src == -1)
			exit_failure("socket: %s", ERRSTR);

		child_pid = fork();
		
		if (child_pid == 0) {
			/* child process */
			close(s_wld);
			play_child(s_src, &srcaddr6.sin6_addr, service);
			/* never reached */
		} else {
			/* parent process */
			close(s_src);
			if (child_pid == -1)
				syslog(LOG_ERR, "can't fork");
		}
	}
	/* never reached */
}

void
play_child(int s_src, struct in6_addr *srcsin6, const char *service)
{
	struct sockaddr_in6 dstaddr6; 
	struct sockaddr_in  dstaddr;
	struct in6_addr faith_prefix;
	char ntop_buf[INET6_ADDRSTRLEN];
	int len = sizeof(dstaddr6);
	int s_dst, err, hport, nresvport, on = 1;
	
	syslog(LOG_INFO,
	       "accepted a client from %s",
	       inet_ntop(AF_INET6, srcsin6, ntop_buf, sizeof(ntop_buf)));

	err = getsockname(s_src, (struct sockaddr *)&dstaddr6, &len);
	if (err == -1)
		exit_failure("getsockname: %s", ERRSTR);

	syslog(LOG_INFO,
	       "the client is connecting to %s",
	       inet_ntop(AF_INET6, &dstaddr6.sin6_addr,
			 ntop_buf, sizeof(ntop_buf)));
	
	{
		int mib[4], size;

		mib[0] = CTL_NET;
		mib[1] = PF_INET6;
		mib[2] = IPPROTO_IPV6;
		mib[3] = IPV6CTL_FAITH_PREFIX;
		size = sizeof(struct in6_addr);
		if (sysctl(mib, 4, &faith_prefix, &size, NULL, 0) < 0) {
			exit_error("sysctl: %s", ERRSTR);
		}
	}
	if (! (dstaddr6.sin6_addr.s6_addr32[0] == faith_prefix.s6_addr32[0] &&
	       dstaddr6.sin6_addr.s6_addr32[1] == faith_prefix.s6_addr32[1] &&
	       dstaddr6.sin6_addr.s6_addr32[2] == faith_prefix.s6_addr32[2])) {
		if (serverpath) {
			/*
			 * Local service
			 */
			syslog(LOG_INFO, "executing local %s", serverpath);
			dup2(s_src, 0);
			close(s_src);
			dup2(0, 1);
			dup2(0, 2);
			execv(serverpath, serverarg);
			syslog(LOG_ERR, "execv %s: %s", serverpath, ERRSTR);
			_exit(EXIT_FAILURE);
		} else {
			close(s_src);
			exit_success("no local service for %s", service);
		}
	}

	/*
	 * Act as a translator
	 */

	syslog(LOG_INFO, "translating from v6 to v4");
	
	memset(&dstaddr, 0, sizeof(dstaddr));
	dstaddr.sin_len = sizeof(dstaddr);
	dstaddr.sin_family = AF_INET;
	dstaddr.sin_port = dstaddr6.sin6_port;
	dstaddr.sin_addr.s_addr = dstaddr6.sin6_addr.s6_addr32[3];

	syslog(LOG_INFO, 
	       "the translator is connecting to %s",
	       inet_ntop(AF_INET, &dstaddr.sin_addr,
			 ntop_buf, sizeof(ntop_buf)));

	hport = ntohs(dstaddr.sin_port);

	switch (hport) {
	case RLOGIN_PORT:
	case RSH_PORT:
		s_dst = rresvport(&nresvport);
		break;
	default:
		s_dst = socket(AF_INET, SOCK_STREAM, 0);
		break;
	}
	if (s_dst == -1)
		exit_failure("socket: %s", ERRSTR);

	err = setsockopt(s_dst, SOL_SOCKET, SO_OOBINLINE, &on, sizeof(on));
	if (err == -1)
		exit_error("setsockopt: %s", ERRSTR);

	err = connect(s_dst, (struct sockaddr *)&dstaddr, sizeof(dstaddr));
	if (err == -1)
		exit_failure("connect: %s", ERRSTR);

	switch (hport) {
	case FTP_PORT:
		ftp_relay(s_src, s_dst);
		break;
	case RSH_PORT:
		rsh_relay(s_src, s_dst);
		break;
	default:
		tcp_relay(s_src, s_dst, service);
		break;
	}

	/* never reached */
}

void
sig_child(int sig)
{
	int status;
	int pid;

	pid = wait3(&status, WNOHANG, (struct rusage *)0);
	if (pid && status)
		syslog(LOG_WARNING, "child %d exit status 0x%x", pid, status);
}

void
sig_terminate(int sig)
{
	syslog(LOG_INFO, "Terminating faith daemon");	
	exit(EXIT_SUCCESS);
}

void
start_daemon(void)
{
	int pid, fd;
	
	if (getppid() == 1) goto cleanup;

	signal(SIGTTOU, SIG_IGN);
	signal(SIGTTIN, SIG_IGN);
	signal(SIGTSTP, SIG_IGN);

	if ((pid = fork()) < 0)
		exit_error("fork: %s", ERRSTR);
	else if (pid > 0)
		exit(EXIT_SUCCESS);

	/* I'm a child */

	if (setpgrp(0, getpid()) == -1)
		syslog(LOG_ERR, "getpid");

	if ((fd = open("/dev/tty", O_RDWR)) >= 0) {
		ioctl(fd, TIOCNOTTY, NULL);
		close(fd);
	}

 cleanup:
	chdir("/");
	umask(0);

	if (signal(SIGCHLD, sig_child) == SIG_ERR)
		exit_failure("signal CHLD: %s", ERRSTR);

	if (signal(SIGTERM, sig_terminate) == SIG_ERR)
		exit_failure("signal TERM: %s", ERRSTR);
}

void
exit_error(const char *fmt, ...)
{
	va_list ap;
	char buf[BUFSIZ];

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);
	fprintf(stderr, "%s\n", buf);
	exit(EXIT_FAILURE);
}

void
exit_failure(const char *fmt, ...)
{
	va_list ap;
	char buf[BUFSIZ];

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);
	syslog(LOG_ERR, fmt);
	exit(EXIT_FAILURE);
}

void
exit_success(const char *fmt, ...)
{
	va_list ap;
	char buf[BUFSIZ];

	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);
	syslog(LOG_INFO, fmt);
	exit(EXIT_SUCCESS);
}
