/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <netinet/ip6.h>
#include <netinet/icmp6.h>

#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <syslog.h>

#define LL_ALLROUTERS "ff02::2"
#define SL_ALLROUTERS "ff05::2"

#define DEF_VLTIME 2400
#define DEF_PLTIME 600

#ifndef IN6_IS_SCOPE_LINKLOCAL
#define IN6_IS_SCOPE_LINKLOCAL(a)	\
	((IN6_IS_ADDR_LINKLOCAL(a)) ||	\
	 (IN6_IS_ADDR_MC_LINKLOCAL(a)))
#endif /* IN6_IS_SCOPE_LINKLOCAL */

struct msghdr sndmhdr;
struct msghdr rcvmhdr;
struct sockaddr_in6 from;
struct sockaddr_in6 sin6_ll_allrouters;

int s;
char *ifname; /* ADHOC */
struct in6_addr prefix; /* ADHOC */
int prefixlen = 64; /* ADHOC */


/* Print usage. Don't call this after daemonized. */
static void
show_usage()
{
	fprintf(stderr, "usage: rrenumd"
/*		" [-c conf_file]" */ /* notyet */
		" -df"
		" interface prefix"	/* ADHOC */
		"\n");
	exit(1);
}

void
init_sin6(struct sockaddr_in6 *sin6, const char *addr_ascii)
{
	memset(sin6, 0, sizeof(*sin6));
	sin6->sin6_len = sizeof(*sin6);
	sin6->sin6_family = AF_INET6;
	inet_pton(AF_INET6, addr_ascii, &sin6->sin6_addr);
}

void
join_multi(const char *addrname)
{
	struct ipv6_mreq mreq;

	if (inet_pton(AF_INET6, addrname, &mreq.ipv6mr_multiaddr.s6_addr)
	    != 1) {
		syslog(LOG_ERR, "<%s> inet_pton failed(library bug?)",
		       __FUNCTION__);
		exit(1);
	}
	/* ADHOC: currently join only one */
	{
		if ((mreq.ipv6mr_interface = if_nametoindex(ifname)) == 0) {
			syslog(LOG_ERR, "<%s> ifname %s should be invalid: %s",
			       __FUNCTION__, ifname, strerror(errno));
			exit(1);
		}
		if (setsockopt(s, IPPROTO_IPV6, IPV6_JOIN_MEMBERSHIP,
			       &mreq,
			       sizeof(mreq)) < 0) {
			syslog(LOG_ERR, "<%s> IPV6_JOIN_MEMBERSHIP on %s: %s",
			       __FUNCTION__, ifname, strerror(errno));
			exit(1);
		}
	}
}

void
init_globals()
{
	static struct iovec rcviov;
	static struct iovec sndiov;
	static u_char sprdata[IPV6_MMTU];
	static u_char rprdata[4500]; /* maximal MTU of connected links */
	static u_char rcvcmsgbuf[CMSG_SPACE(sizeof(struct in6_pktinfo)) +
				 CMSG_SPACE(sizeof(int))];
	static u_char sndcmsgbuf[CMSG_SPACE(sizeof(struct in6_pktinfo)) +
				 CMSG_SPACE(sizeof(int))];

	/* init ll_allrouters */
	init_sin6(&sin6_ll_allrouters, LL_ALLROUTERS);

	/* initialize msghdr for receiving packets */
	rcviov.iov_base = (caddr_t)rprdata;
	rcviov.iov_len = sizeof(rprdata);
	rcvmhdr.msg_namelen = sizeof(struct sockaddr_in6);
	rcvmhdr.msg_iov = &rcviov;
	rcvmhdr.msg_iovlen = 1;
	rcvmhdr.msg_control = (caddr_t)rcvcmsgbuf;
	rcvmhdr.msg_controllen = sizeof(rcvcmsgbuf);

	/* initialize msghdr for sending packets */
	sndiov.iov_base = (caddr_t)sprdata;
	sndmhdr.msg_namelen = sizeof(struct sockaddr_in6);
	sndmhdr.msg_iov = &sndiov;
	sndmhdr.msg_iovlen = 1;
	sndmhdr.msg_control = (caddr_t)sndcmsgbuf;
	sndmhdr.msg_controllen = sizeof(sndcmsgbuf);
}

void
config()
{
	/* TODO; parse config file */
	/* ADHOC: just make ADD message for arg prefix */
	struct icmp6_hdr *icp;
	struct icmp6_router_renum *rr;
	struct rr_pco_match *mp;
	struct rr_pco_use *up;

	memset(sndmhdr.msg_iov->iov_base, 0,
	       sizeof(struct icmp6_hdr) + sizeof(struct rr_pco_use));
	rr = (struct icmp6_router_renum *)sndmhdr.msg_iov->iov_base;

	rr->rr_type = ICMP6_ROUTER_RENUMBERING;
	rr->rr_code = 0;
	rr->rr_cksum = 0;
	rr->rr_seqnum = 0;
	rr->rr_segnum = 0;
	rr->rr_test = 0;
	rr->rr_reqresult = 0;
	rr->rr_forceapply = 0;
	rr->rr_specsite = 1;
	rr->rr_prevdone = 0;
	rr->rr_maxdelay = 0; /* ADHOC */
	rr->rr_authlen = 0; /* ADHOC */
	rr->rr_authoff = htons(sizeof(*rr) + sizeof(*mp)
			 	+ sizeof(*up)); /* ADHOC */
	rr->rr_keyid = htons(0); /* ADHOC */

	mp = (struct rr_pco_match *)(rr + 1);
	mp->rpm_code = RPM_PCO_ADD; /* ADHOC */
	mp->rpm_len = 4*1/* 1 use_prefix */ + 3; /* as router-renum-04.txt */
	mp->rpm_ordinal = 0; /* ADHOC */
	mp->rpm_matchlen = 0; /* ADHOC */
	mp->rpm_reserved = 0;
	mp->rpm_prefix = in6addr_any; /* ADHOC */

	up = (struct rr_pco_use *)(mp + 1);
	up->rpu_uselen = 64; /* ADHOC */
	up->rpu_keeplen = 0; /* ADHOC */
	up->rpu_useorig_onlink = 0;
	up->rpu_useorig_autonomous = 0;
	up->rpu_useorig_reserved = 0;
	up->rpu_onlink = 1;
	up->rpu_autonomous = 1;
	up->rpu_raflags_reserved = 0;
	up->rpu_vltime = htonl(DEF_VLTIME);
	up->rpu_pltime = htonl(DEF_PLTIME);
	up->rpu_decr_vltime = 1;
	up->rpu_decr_pltime = 1;
	up->rpu_flags_reserved = 0;
	up->rpu_reserved = 0;
	up->rpu_prefix = prefix; /* ADHOC */

	sndmhdr.msg_iov->iov_len = sizeof(*rr) + sizeof (*mp)
		+ sizeof(*up); /* ADHOC */
}

void
sock_open()
{
	struct icmp6_filter filt;
	int on;

	if ((s = socket(AF_INET6, SOCK_RAW, IPPROTO_ICMPV6)) < 0) {
		syslog(LOG_ERR, "<%s> socket: %s", __FUNCTION__,
		       strerror(errno));
		exit(1);
	}

	/*
	 * join all routers multicast addresses.
	 */
	join_multi(LL_ALLROUTERS);
	join_multi(SL_ALLROUTERS);

	/* set icmpv6 filter */
	ICMP6_FILTER_SETBLOCKALL(&filt);
	ICMP6_FILTER_SETPASS(ICMP6_ROUTER_RENUMBERING, &filt);
	if (setsockopt(s, IPPROTO_ICMPV6, ICMP6_FILTER, &filt,
		       sizeof(filt)) < 0) {
		syslog(LOG_ERR, "<%s> IICMP6_FILTER: %s",
		       __FUNCTION__, strerror(errno));
		exit(1);
	}

	/* specify to tell receiving interface */
	on = 1;
	if (setsockopt(s, IPPROTO_IPV6, IPV6_PKTINFO, &on,
		       sizeof(on)) < 0) {
		syslog(LOG_ERR, "<%s> IPV6_PKTINFO: %s",
		       __FUNCTION__, strerror(errno));
		exit(1);
	}

	return;
}

void
rrenum_output(struct sockaddr_in6 *sin6, char *ifname)
{
	int i;
	static int seqnum = 0;
	struct cmsghdr *cm;
	struct in6_pktinfo *pi;
	struct icmp6_router_renum *rr;

	sndmhdr.msg_name = (caddr_t)sin6;

	cm = CMSG_FIRSTHDR(&sndmhdr);
	/* specify the outgoing interface */
	cm->cmsg_level = IPPROTO_IPV6;
	cm->cmsg_type = IPV6_PKTINFO;
	cm->cmsg_len = CMSG_LEN(sizeof(struct in6_pktinfo));
	pi = (struct in6_pktinfo *)CMSG_DATA(cm);
	memset(&pi->ipi6_addr, 0, sizeof(pi->ipi6_addr));	/*XXX*/
	pi->ipi6_ifindex = if_nametoindex(ifname);

	/* specify the hop limit of the packet if dest is link local */
	/* not defined by router-renum-04.txt, but maybe its OK */
	if (IN6_IS_SCOPE_LINKLOCAL(&sin6->sin6_addr)) {
		int hoplimit = 255;

		cm = CMSG_NXTHDR(&sndmhdr, cm);
		cm->cmsg_level = IPPROTO_IPV6;
		cm->cmsg_type = IPV6_HOPLIMIT;
		cm->cmsg_len = CMSG_LEN(sizeof(int));
		memcpy(CMSG_DATA(cm), &hoplimit, sizeof(int));
	}

	/* ADHOC: update seqnum */
	rr = (struct icmp6_router_renum *)sndmhdr.msg_iov->iov_base;
	rr->rr_seqnum = htonl(seqnum);
	seqnum++;

	i = sendmsg(s, &sndmhdr, 0);

	if (i < 0 || i != sndmhdr.msg_iov->iov_len)
		syslog(LOG_ERR, "<%s> sendmsg: %s", __FUNCTION__,
		       strerror(errno));
}

void
rrenum_input()
{
	int i;
	struct icmp6_router_renum *rr;

	/* get message */
	if ((i = recvmsg(s, &rcvmhdr, 0)) < 0) {
		syslog(LOG_ERR, "<%s> recvmsg: %s", __FUNCTION__,
		       strerror(errno));
		return;
	}

	if (i < sizeof(struct icmp6_router_renum)) {
		syslog(LOG_ERR, "<%s> packet size(%d) is too short",
		       __FUNCTION__, i);
		return;
	}

	rr = (struct icmp6_router_renum *)rcvmhdr.msg_iov->iov_base;

	switch(rr->rr_code) {
	case ICMP6_ROUTER_RENUMBERING_COMMAND:
		/* COMMAND will be processed by rtadvd */
		break;
	case ICMP6_ROUTER_RENUMBERING_RESULT:
		/* TODO: receiving result message */
		break;
	default:
		syslog(LOG_ERR,	"<%s> received unknown code %d"
		       __FUNCTION__, rr->rr_code);
		break;
	}
}

int
main(int argc, char *argv[])
{
	char *cfile = NULL;
	fd_set fdset;
	struct timeval timeout;
	int ch, i;
	int dflag = 0, fflag = 0;
	int maxfd = 0;

	openlog(*argv, LOG_PID, LOG_DAEMON);

	/* get options */
	while ((ch = getopt(argc, argv, "cdf")) != -1){
		switch (ch) {
		case 'c':
			cfile = optarg;
			break;
		case 'd':
			dflag = 1;
			break;
		case 'f':
			fflag = 1;
			break;
		default:
			show_usage();
		}
	}
	argc -= optind;
	argv += optind;

	/* set log level */
	if (dflag == 0)
		(void)setlogmask(LOG_UPTO(LOG_ERR));
	if (dflag == 1)
		(void)setlogmask(LOG_UPTO(LOG_INFO));

	if (argc == 2) { /* ADHOC */
		ifname = *argv; /* ifname will be checked in join_multi() */
		if (inet_pton(AF_INET6, *(argv + 1), &prefix) != 1) {
			syslog(LOG_ERR,
			       "<%s> inet_pton failed. prefix $s should be wrong",
			       __FUNCTION__, *(argv + 1));
			exit(1);
		}
	} else
		show_usage();

	/* init global variables */
	init_globals();

	config();

	sock_open();

	if (!fflag)
		daemon(0, 0);

	FD_ZERO(&fdset);
	FD_SET(s, &fdset);
	maxfd = s;

	/* ADHOC: timeout each 30seconds */
	memset(&timeout, 0, sizeof(timeout));
	timeout.tv_sec = 30;

	while (1) {
		struct fd_set select_fd = fdset; /* reinitialize */

		if ((i = select(maxfd + 1, &select_fd, NULL, NULL,
				&timeout)) < 0){
			syslog(LOG_ERR, "<%s> select: %s",
			       __FUNCTION__, strerror(errno));
			continue;
		}
		if (i == 0) {	/* timeout */
			/* send to each direct interfaces */
			/* ADHOC: only send to one */
			rrenum_output(&sin6_ll_allrouters, ifname);
			/* TODO; send to site local all routers */
		}
		if (FD_ISSET(s, &select_fd))
			rrenum_input();
	}
}
