/*
 * Copyright (C) 1995, 1996, 1997, and 1998 WIDE Project.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *    This product includes software developed by WIDE Project and
 *    its contributors.
 * 4. Neither the name of the project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE PROJECT AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE PROJECT OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/param.h>
#include <sys/socket.h>

#include <net/if.h>
#include <netinet/in.h>
#include <netinet/icmp6.h>

#include <arpa/inet.h>

#include <syslog.h>

#define RR_MAX_KEYNUM	10 /* ADHOC */

#define	RR_ISSET_SEGNUM(segnum_bits, segnum) \
	((((segnum_bits)[(segnum) >> 5]) & (1 << ((segnum) & 31))) != 0)
#define	RR_SET_SEGNUM(type, filterp) \
	(((segnum_bits)[(segnum) >> 5]) |= (1 << ((segnum) & 31)))

struct rr_operation {
	u_short	rro_keyid;
	u_long	rro_seqnum;
	u_long	rro_segnum_bits[8];
};

struct rr_operation rro[RR_MAX_KEYNUM];
u_char rro2key[RR_MAX_KEYNUM];

void
do_rr()
{
}

/* return 0 or positive value on success, -1 on failure */
static int
get_rro_index(int keyid)
{
	int i;

	/* rro_index 0 is reserved for keyid 0 */
	if (keyid == 0)
		return 0;

	/* if rro_index != 0, also keyid != 0 */
	for (i = 1; i < RR_MAX_KEYNUM; i++) {
		if (rro2key[i] == 0) {
			rro2key[i] = keyid;
			return i;
		}
		else if (rro2key[i] == keyid)
			return i;
	}

	return -1;
}

/* return 0 on success, 1 on failure */
static int
set_segnum(u_long segnum_bits[], u_char segnum)
{
	if (RR_ISSET_SEGNUM(segnum_bits, segnum))
		return 1;
	RR_SET_SEGNUM(segnum_bits, segnum);
	return 0;
}

/* return 0 on success, 1 on failure */
static int
rr_command_check(struct icmp6_router_renum *rr, struct in6_addr *from,
		 struct in6_addr *dst)
{
	u_char ntopbuf[INET6_ADDRSTRLEN];
	int rro_index;

	/* omit minimal length check. hope kernel have done it. */

	/* destination check. only for multicast. omit unicast check. */
	if (IN6_IS_ADDR_MULTICAST(dst) && !IN6_IS_ADDR_MC_LINKLOCAL(dst) &&
	    !IN6_IS_ADDR_MC_SITELOCAL(dst)) {
		syslog(LOG_ERR,	"<%s> dst mcast addr %s is illegal"
		       __FUNCTION__,
		       inet_ntop(AF_INET6, dst, ntopbuf, INET6_ADDRSTRLEN));
		return 1;
	}

	/* get rro_index */
	rro_index = get_rro_index(rr->rr_keyid);
	if (rro_index < 0) {
		syslog(LOG_ERR,
		       "<%s> get rro_id err for %d. maybe reached MAX KEYNUM",
		       __FUNCTION__, ntohl(rr->rr_seqnum));
		return 1;
	}

	/* seqnum and segnum check */
	if (rro[rro_index].rro_seqnum > rr->rr_seqnum) {
		syslog(LOG_WARNING,
		       "<%s> rcvd old seqnum %d from %s",
		       __FUNCTION__, ntohl(rr->rr_seqnum),
		       inet_ntop(AF_INET6, from, ntopbuf, INET6_ADDRSTRLEN));
		return 1;
	}
	if (rro[rro_index].rro_seqnum == rr->rr_seqnum &&
	    rr->rr_test == 0 &&
	    set_segnum(rro[rro_index].rro_segnum_bits, rr->rr_segnum)) {
		if (rr->rr_reqresult)
			syslog(LOG_WARNING,
			       "<%s> rcvd duped segnum %d from %s",
			       __FUNCTION__, rr->rr_segnum,
			       inet_ntop(AF_INET6, from, ntopbuf,
					 INET6_ADDRSTRLEN));
		return 0;
	}

	/* TODO security check. should delete illegal keyid from rro2key[] */

	/* update seqnum */
	rro[rro_index].rro_seqnum = rr->rr_seqnum;

	return 0;
}

static void
rr_command_input(int len, struct icmp6_router_renum *rr,
		 struct in6_addr *from, struct in6_addr *dst, int rcvifindex)
{
	if (rr_command_check(rr, from, dst)) {
		syslog(LOG_ERR, "<%s> received RR was invalid", __FUNCTION__);
		return;
	}
	if (rr->rr_test)
		return;

	/* do router renumbering */
	/* ADHOC: now only for rcv interface */
	do_rr(rcvifindex, rr);

	return;
}

void
rr_input(int len, struct icmp6_router_renum *rr, struct in6_pktinfo *pi,
	 struct sockaddr_in6 *from, struct in6_addr *dst)
{
	u_char ntopbuf[2][INET6_ADDRSTRLEN], ifnamebuf[IFNAMSIZ];

	syslog(LOG_DEBUG,
	       "<%s> RR received from %s to %s on %s",
	       __FUNCTION__,
	       inet_ntop(AF_INET6, &from->sin6_addr,
			 ntopbuf[0], INET6_ADDRSTRLEN),
	       inet_ntop(AF_INET6, &dst, ntopbuf[1], INET6_ADDRSTRLEN),
	       if_indextoname(pi->ipi6_ifindex, ifnamebuf));

	/* TODO: some consistency check. */

	switch (rr->rr_code) {
	case ICMP6_ROUTER_RENUMBERING_COMMAND:
		rr_command_input(len, rr, &from->sin6_addr, dst,
				 pi->ipi6_ifindex);
		/* TODO: send reply msg */
	case ICMP6_ROUTER_RENUMBERING_RESULT:
		/* RESULT will be processed by rrenumd */
		break;
	default:
		syslog(LOG_ERR,	"<%s> received unknown code %d"
		       __FUNCTION__, rr->rr_code);
		break;

	}

	return;
}
